/* This demo program plays a simple game involving aligning colored blocks
to get rows of three and thus disappear scoring points.

It is designed to show how classes can actually be used in the base design of
a program and not just for libraries.

Each class defined has a comment to show any interesting features contained
within
*/



#include <textwin.hpp>
#include <stdlib.h>
#include <string.h>
#ifdef __OS2__
#include "OS2Glue.hpp"
#else
#include <bios.h>
#endif
#include "highscr.hpp"

window whole(0,0,80,25,Black,Black);

#ifdef __OS2__
#ifdef _mthread
#error The keyboard routines will not work in MThread mode
#endif
#endif

/* The following very tiny class is to allow you to specify a state for
the lockable keys (namely caps, num, scroll and insert) for the duration
of a class scope. Note that the status is set on construction and reset
on exit and that no attempt is made to enforce the status during keypress
sequences. IE If the application user wants to upset things (s)he can */

#define InsertLocked (0x80)
#define CapsLocked (0x40)
#define NumLocked (0x20)
#define ScrollLocked (0x10)

class keystate
{ unsigned char oldstatus;
  public:
    keystate(unsigned char);
    ~keystate();
};

keystate::keystate(unsigned char x)
{
#ifdef __OS2__
#else
  unsigned char far *statusbyte=(unsigned char far*)0x00400017;
  oldstatus=*statusbyte;
  *statusbyte=oldstatus&0xF|x;
#endif
}

keystate::~keystate()
{
  #ifdef __OS2__
  #else
  unsigned char far*statusbyte=(unsigned char far*)0x00400017;
  *statusbyte=*statusbyte&0xF|oldstatus;
  #endif
}

/* This class is underused by the program but is included for completeness.
The aim is to encapsulate all timing within one area of code so that the speed
of the program can be controlled centrally and consistantly.
*/

enum delay_length {long_delay,medium_delay,short_delay};
class Pause
{ private:
    int lengths[3];
    int minimums[3];
    void check_range();
  public:
    Pause(int m1,int l1,int m2,int l2,int m3,int l3); // Longest first
    void operator*=(int percent);
    void operator()(delay_length which)
    { delay(lengths[char(which)]);
    }
};

Pause::Pause(int m1,int l1,int m2,int l2,int m3,int l3)
{ lengths[0]=l1;
  lengths[1]=l2;
  lengths[2]=l3;
  minimums[0]=m1;
  minimums[1]=m2;
  minimums[2]=m3;
  check_range();
}

void Pause::operator*=(int percent)
{ for (char i=0;i<3;i++)
    lengths[i]=int((long(lengths[i])*percent)/100);
  check_range();
}

void Pause::check_range()
{ for (char i=0;i<3;i++)
    if (lengths[i]<minimums[i])
      lengths[i]=minimums[i];
}

Pause wait(5,1500,5,250,2,50);

/* The DisplayedInt class is really just an integer together with a window in
which to display it. The interesting feature is that the instances of this
class hook themselves onto an activity chain which can be called by the program
at 'idle moments'. This might prove handy if a large interactive display was
being maintained. It might have been neater to have defined the chaining
facility in a base class to the all displayed items could be derived from this */

class DisplayedInt
{ private:
    int value;
    window *win;
    DisplayedInt *next;
  public:
    static DisplayedInt *activitychain;
    void Activate();
    void operator=(int x)
    { value=x;
    }
    DisplayedInt(char *title,Color fore,Color back,int x,int y,int startval=0);
    void operator+=(int x)
    { value+=x;
    }
    int operator()()
    { return value;
    }
    ~DisplayedInt()
    { delete win;
    }
};
DisplayedInt *DisplayedInt::activitychain=0;

void DisplayedInt::Activate()
{ win->use();
  clrscr();
  char temp[10];
  itoa(value,temp,10);
  char len=strlen(temp);
  for (int i=0;i<(10-len)/2;i++)
    putch(' ');
  cputs(temp);
  if (next)
    next->Activate();
}

DisplayedInt::DisplayedInt(char *title,Color fore,Color back,int x,int y,int startval)
{ win=new window(x,y,x+10,y+1,DOUBLEFRAME,fore,back,fore,back);
  win->settitle(title,CenterUpperTitle);
  cursoroff();
  value=startval;
  next=activitychain;
  activitychain=this;
}

DisplayedInt Score("Score",White,Green,65,5,0);
DisplayedInt Blocks("Blocks",White,Red,65,11,0);
DisplayedInt Level("Level",Black,Cyan,65,17,0);

/* The program revolves around groups of three blocks. These are therefore
a class so that rules about thier use are contained */

class Tile
{ private:
    Color data[3];
    Color newtile();
  public:
    void RFill()
    { data[0]=newtile();
      data[1]=newtile();
      data[2]=newtile();
    }
    Tile(Color x,Color y,Color z)
    { data[0]=x;
      data[1]=y;
      data[2]=z;
    }
    Tile()
    { RFill();
    }
    void operator=(Tile &x)
    { memmove(this->data,x.data,sizeof(int)*3);
    }
    Tile(Tile &x)
    { memmove(this->data,x.data,sizeof(int)*3);
    }
    void Rotate()
    { Color t=data[0];
      data[0]=data[1];
      data[1]=data[2];
      data[2]=t;
    }
    Color& operator[](char x)
    { return data[x];
    }
    char Score()
    { if (data[0]==LightGray) return 0;
      return data[0]==data[1] && data[1]==data[2];
    }
};

Color Tile::newtile()
{ Color t=Color(1+rand()%6);
  if (t==Brown)
    t=Yellow;
  if (t==Cyan)
    t=LightCyan;
  return t;
}

/* The main playing area is a single grid. All rules regarding current
state of play are therefore contained in this class */

enum out_event { touched_bottom,full,nothing };
enum in_event { rotate,left,right,drop,fall,nout };

class Grid
{ public:
    Grid();
    ~Grid();
    out_event Enter();
    out_event Drop1(in_event move);
    char Reduce();
    char Dropped()
    { return lastdrop;
    }
  private:
    Color table[6][18];
    Tile CurTile;
    char CurCol,CurY;
    char lastdrop;
    void placeblock(char x,char y,Color c);
    void placetile(in_event e);
    window *grid;
};

Grid::Grid()
{ grid=new window(34,1,45,18,LightGray,LightGray);
  cursoroff();
  setwrap(0);
  memwset(table,LightGray,6*18);
}

Grid::~Grid()
{ delete grid;
}

void Grid::placeblock(char x,char y,Color c)
{ grid->use();
  gotoxy(2*x+1,18-y);
  textcolor(c);
  putch('\xDB');
  putch('\xDB');
  table[x][y]=c;
}

void Grid::placetile(in_event e)
{ char i;
  if (e==fall)
    placeblock(CurCol,CurY+3,LightGray);
  else
    if (e==left || e==right)
      for (i=0;i<3;i++)
        placeblock(CurCol+(e==left?1:-1),CurY+i,LightGray);
    else
      if (e==drop)
        for (i=CurY+3;i<18;i++)
          placeblock(CurCol,i,LightGray);
  for (i=0;i<3;i++)
    placeblock(CurCol,CurY+i,CurTile[i]);
}

out_event Grid::Enter()
{ CurTile.RFill();
  CurCol=3;
  CurY=15;
  Color woz=table[3][15];
  placetile(nout);
  return woz==LightGray?nothing:full;
}

out_event Grid::Drop1(in_event e)
{ if (e==left && (!CurCol || table[CurCol-1][CurY]!=LightGray))
    return nothing;
  if (e==right && (CurCol==5 || table[CurCol+1][CurY]!=LightGray))
    return nothing;
  if (e==rotate)
    CurTile.Rotate();
  lastdrop=0;
  if (e==drop)
  { while (CurY && table[CurCol][CurY-1]==LightGray)
    { CurY--;
      lastdrop++;
    }
  }
  else
    if (e==fall)
      CurY--;
  if (e==left)
    CurCol--;
  if (e==right)
    CurCol++;
  if (e==fall && (CurY<0 || table[CurCol][CurY]!=LightGray))
    return touched_bottom;
  placetile(e);
  return e==drop?touched_bottom:nothing;
}

char Grid::Reduce()
{ char spare[6][18];
  wait(medium_delay);
  memset(spare,0,6*18);
  for (char i=0;i<18;i++)
    for (char j=0;j<4;j++)
      if (Tile(table[j][i],table[j+1][i],table[j+2][i]).Score())
        spare[j][i]=spare[j+1][i]=spare[j+2][i]=1;
  for (i=0;i<16;i++)
    for (j=0;j<6;j++)
      if (Tile(table[j][i],table[j][i+1],table[j][i+2]).Score())
        spare[j][i]=spare[j][i+1]=spare[j][i+2]=1;
  for (i=0;i<16;i++)
    for (j=0;j<4;j++)
      if (Tile(table[j][i],table[j+1][i+1],table[j+2][i+2]).Score())
        spare[j][i]=spare[j+1][i+1]=spare[j+2][i+2]=1;
  for (i=0;i<16;i++)
    for (j=2;j<6;j++)
      if (Tile(table[j][i],table[j-1][i+1],table[j-2][i+2]).Score())
        spare[j][i]=spare[j-1][i+1]=spare[j-2][i+2]=1;
  for (i=0;i<18;i++)
    for (j=0;j<6;j++)
      if (spare[j][i])
        placeblock(j,i,Black);
  wait(medium_delay);
  char total=0;
  for (j=0;j<6;j++)
  { char CY=0;
    for (i=0;i<18;i++)
    { while (table[j][i]==Black)
      { total++;
        i++;
      }
      if (table[j][i]==LightGray)
        break;
      if (CY!=i)
        placeblock(j,CY,table[j][i]);
      CY++;
    }
    while (CY<18 && table[j][CY]!=LightGray)
      placeblock(j,CY++,LightGray);
  }
  return total?total+Reduce():total;
}

Grid Board;

/* With all the active items defined it simply remains to integrate the
keyboard input from the user */

void main_loop()
{ out_event oe=Board.Enter();
  char w=0;
  while (oe!=full)
  { in_event ev=fall;
    int key=0;
    for (w=w>19?0:w;w<20 && !key;w++)
    { key=bioskey(1);
      wait(short_delay);
    }
    if (w<20)
    { int key=bioskey(0);
      switch (char(key>>8))
      { case 0x4B:  // Left
          ev=left;
          break;
        case 0x4D:  // Right
          ev=right;
          break;
        case 0x4C:  // Up or numeric 5 key
        case 0x48:
          ev=rotate;
          break;
        case 0x50: //Down
          ev=drop;
          break;
        case 0x10:    //Q
          return;
      }
    };
    if (Board.Drop1(ev)==touched_bottom)
    { char got=Board.Reduce();
      Blocks+=got;
      Score+=got*(1+Board.Dropped()/5)*(1+Level()/3);
      if (Blocks()/50>Level())
      { Level=Blocks()/50;
        wait*=90;
      }
      DisplayedInt::activitychain->Activate();
      oe=Board.Enter();
    }
  }
}

window RuleWindow(0,0,20,10,DOUBLEFRAME,LightGray,Blue,LightGray,Blue);

void RulesScreen()
{ RuleWindow.use();
  cputs("Get as many lines \nof three colours asyou can before \nfilling the grid\n\n");
  cputs("\x1a\x1b  Move tile\n");
  cputs("\x19   Drop Tile\n");
  cputs("\x18   Rotate Colors\n");
  cputs("Q   Quit Game");
  cursoroff();
}

/* Note that the highscore is maintained by a seperate and (hopefully)
reusable module */

keystate keys(NumLocked);

main()
{ HighScoreTable winners("column.hig",15);
  whole.use();
  clrscr();
  RulesScreen();
  winners.Display();
  randomize();
  main_loop();
  winners.AddValue(Score());
  winners.Display();
}