// TopSpeed C++ demonstration program
// This simple traffic simulation illustrates the use of multi-threading,
// as well as many other features of the TopSpeed C++ language

// This program should be made using the supplied TRAFFIC.PR project file

#pragma check(stack=>off,index=>off,nil_ptr=>off)

#ifndef _mthread
  #error This program is only valid in multithread model
#endif

#include <process.h>
#include <textwin.hpp>
#include <coniostr.hpp>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>



int randto(int upper)
{ return rand() % upper;
}

const int MAXCAR = 100;
const int xNum   = SCREENWIDTH;
const int yNum   = SCREENDEPTH;

unsigned ToNorth[yNum], ToSouth[yNum], ToEast[xNum], ToWest[xNum];

#ifdef __OS2__
const int IsBW = 0;
#else

int CheckBW() =
{
  0xB4, 0x0F,           //     mov   ah,0FH
  0xCD, 0x10,           //     int   10H

  0x0A, 0xC0,           //     or    al,al
  0x74, 0x12,           //     je    $1
  0x3C, 0x02,           //     cmp   al,2
  0x74, 0x0E,           //     je    $1
  0x3C, 0x05,           //     cmp   al,5
  0x74, 0x0A,           //     je    $1
  0x3C, 0x06,           //     cmp   al,6
  0x74, 0x06,           //     je    $1
  0x3C, 0x07,           //     cmp   al,7
  0x74, 0x02,           //     je    $1
                        //  $0:
  0x33, 0xC0,           //     xor ax,ax
                        //  $1:
  0xCB                  //     ret far 0
};
int IsBW = CheckBW();
#endif

#define MAX_SPEED 450
#define MAX_ACCEL 80

enum Direction { East=1,South=2,North=4,West=8 };
inline Direction Reverse(Direction d)
{ switch (d)
  { case East:  return West;
    case West:  return East;
    case South: return North;
    case North: return South;
  }
}

Direction RandDirection()
{ return Direction(1<<randto(4));
}

class Location {
  friend class Car;
  unsigned char Entry, Exit;  // bits represent ESNW entries or exits
public:
  unsigned char Occupied;     // 0 or 1
  unsigned char Scenery;
  int  IsExit  (unsigned char e) { return e & Exit;  };
  void AddExit (unsigned char e) { Exit |= e;        };
  int  IsEntry (unsigned char e) { return e & Entry; };
  void AddEntry(unsigned char e) { Entry |= e;       };
  void RemoveEntry(unsigned char e) { Entry &= ~e; };
  int  Empty() { return (Exit==0); };
  friend void PlotLoc(int x,int y);
} World[xNum][yNum];

class Car {
  unsigned xpos,ypos;
  Color color;
  unsigned Speed;      // in mph * 10
  Direction LastD;
  Car *Next, *Prev;    // Doubly-linked list
  static Car *Head;    // All active cars
public:
  volatile static unsigned NumCars;
  // member functions
private:
  void Plot(Direction);
public:
  int Move();
  static void Drive();
  ~Car();
  Car();
  friend Car *ChooseCar();
};

volatile unsigned Car::NumCars = 0;
Car* Car::Head = NULL;
Car::Car()
{ int x, y;
  do {
    x = randto(xNum); y = randto(yNum);
  } while (World[x][y].Occupied || World[x][y].Empty());
  Direction d;
  do {
    d = RandDirection();
  } while (!World[x][y].IsExit(d));
  World[x][y].Occupied = TRUE ;
  xpos = x ; ypos = y ;
  Speed = randto(100);
  LastD = d;
  color = IsBW ? White : Color(9+randto(7));
  if ((Next=Head)!=NULL) Head->Prev = this;
  Prev=NULL;
  Head = this;
  NumCars++;
  Plot(d);
}

Car::~Car()
{ int x = xpos, y = ypos ;
  if (NumCars == 0)
    exit(0);
  World[x][y].Occupied = FALSE ;
  PlotLoc(x,y);
//  (Prev==NULL ? Head : Prev->Next)=Next;  // unlink from list
  if (Prev==NULL)
    Head=Next;
  else
    Prev->Next=Next;  // unlink from list
  if (Next!=NULL) Next->Prev = Prev;
  NumCars--;
}

void Car::Plot(Direction d)
{ static /* const */ char CarSym[] = {0,16,31,0,30,0,0,0,17};
  textcolor(color);
  _directwrite(xpos+1,ypos+1,&CarSym[d],1) ;
}

int Car::Move()
{ if (Speed<MAX_SPEED)
  { Speed += randto(MAX_ACCEL);
    if (Speed>MAX_SPEED)
      Speed = MAX_SPEED;
    if (Speed<randto(MAX_SPEED))
      return Speed;
  }
  int x = xpos, y = ypos;
  unsigned char tried = 0;
  while(TRUE)
  { Direction exit = RandDirection();
    if (World[x][y].IsExit(exit))
    { int newx = x, newy = y;
      switch (exit) {
        case East  : newx = ToEast[x];
                     break;
        case South : newy = ToSouth[y];
                     break;
        case North : newy = ToNorth[y];
                     break;
        case West  : newx = ToWest[x];
                     break;
      }
      if (World[newx][newy].Occupied)
      { Speed /= 4; // Brake
        return Speed;
      }
      else
      { World[x][y].Occupied = FALSE ;
        PlotLoc(x,y);
        World[newx][newy].Occupied = TRUE ;
        xpos = newx;
        ypos = newy;
        if (LastD!=exit)
        { // Cornering
          Speed /= 2;
          LastD = exit;
        }
      }
      Plot(exit);
      return Speed;
    }
    tried |= exit;
    if (tried==0xf)  // nowhere to go
    { Speed = 0;
      return Speed;
    }
  }
}

volatile unsigned AvSpeed;   // Must be volatile, as it's used by two processes

void Car::Drive()
{ int up = TRUE;
  while(1) {
    unsigned SumSpeed = 0;
    Car *finger = Head;
    do {
      SumSpeed += finger->Move();
      finger = finger->Next;
    } while (finger!=NULL);
    if (kbhit()) {
      getch();
      break;
    }
    AvSpeed = SumSpeed / NumCars;
    // Add or subtract car
    if (randto(5)==0) {
      if (NumCars==1)
        up = TRUE;
      else if (NumCars==MAXCAR)
        up = FALSE;
      if (up)
        new Car;
      else
        delete  ChooseCar();
      textcolor(Black) ;
    }
  }
}

Car* ChooseCar()
{ // returns a car to be destroyed. This could pick one at random,
  // but as written it always returns the first car.
  return Car::Head;
}

int WrapRoads = 1;

void PlotLoc(int x, int y)
{ static const char Road[] = "   Ⱥ ͻ˼ʹ";
  Location &loc = World[x][y];
  char sc = Road[loc.Entry | loc.Exit];
  if (sc!=32)
  { Color color = (sc=='') ? (IsBW ? LightGray : Brown) : LightGray;
    textcolor(color);
    _directwrite(x+1,y+1,&sc,1);
  }
}

int RoadOk(int x, int y)
// Returns whether we have planning permission to build road at x,y
// Tries to avoid making very small plots of land
{ return (World[x][y].Empty()                  ||
          World[ToEast[x]][y].Empty()          ||
          World[ToEast[x]][ToSouth[y]].Empty() ||
          World[x][ToSouth[y]].Empty());
}

int OnMap(int x, int y)
{ return WrapRoads || (x>1 && y>1 && x<xNum-2 && y<yNum-2);
}

void InitWorld()
{
  for ( int x=0; x<xNum; x++) {
    ToEast[x] = (x+1) % xNum;
    ToWest[x] = (x+xNum-1) % xNum;
  }
  for ( int y=0; y<yNum; y++) {
    ToSouth[y] = (y+1) % yNum;
    ToNorth[y] = (y+yNum-1) % yNum;
  }
  x = xNum / 2;
  y = yNum / 2;
  int newx = x, newy = y;
  Direction newdir = RandDirection();
  int n = 1000 + randto(1000);
  // initialise map using randomish walk of n steps
  while (n > 0 || x != xNum/2 || y != yNum/2) {
    if (n>0) {
      n--;
      if (randto(4)==0)
        newdir = RandDirection();
    }
    else
    { // Get back to the starting point, to ensure no traps
      // Pick a direction that takes us back, unless all such are illegal
      if (x>xNum/2) newdir=West;
      else if (x<xNum/2) newdir=East;
      else if (y>yNum/2) newdir=North;
      else if (y<yNum/2) newdir=South;
      if (randto(4)==0) newdir = RandDirection();
    }
    while (World[x][y].IsEntry(newdir)) // enforce one-way streets
      newdir = RandDirection() ;
    newx = x; newy = y;
    switch (newdir) {
      case East  : newx = ToEast[x];
                   break;
      case South : newy = ToSouth[y];
                   break;
      case North : newy = ToNorth[y];
                   break;
      case West  : newx = ToWest[x];
                   break;
    }
    Direction newentry = Reverse(newdir);
    if (! World[newx][newy].IsEntry(newentry)) {
      World[newx][newy].AddEntry(newentry);
      if (OnMap(newx,newy) &&
          (randto(50)==0 ||
           (RoadOk(newx, newy)                   &&
            RoadOk(ToWest[newx], newy)           &&
            RoadOk(ToWest[newx], ToNorth[newy])  &&
            RoadOk(newx, ToNorth[newy]) ))) {
        World[x][y].AddExit(newdir);
        PlotLoc(x,y);
        PlotLoc(newx,newy);
        PlotLoc(ToEast[x],y);
        PlotLoc(ToWest[x],y);
        PlotLoc(x, ToNorth[y]);
        PlotLoc(x, ToSouth[y]);
        x = newx; y = newy;
      }
      else
        World[newx][newy].RemoveEntry(newentry);
    }
    else {
      x = newx;
      y = newy;
    }
  }
  // make some cars
  for (int v = randto(MAXCAR-1)+1; v--; 0)
    new Car;
}

void Intro()
{ char line[SCREENWIDTH];
  window DescWin(0,11,79,25,DOUBLEFRAME,Black,LightGray);
  cursoroff();
  DescWin.settitle(" JPI TopSpeed C++ : Traffic Simulation ",CenterUpperTitle);
  conout << " This program is a simple simulation of traffic flow in a closed road system.\n";
  conout << " Up to " << MAXCAR << " cars move randomly within a one-way network of roads.\n\n";
  conout << " The average speed is monitored by a separate process and displayed in\n";
  conout << " the Statistics window. 30.0 MPH indicates no congestion.\n\n";
  conout << " Any resemblance to rush-hour traffic in New York is purely coincidental.\n\n\n";
  textcolor(White);
  conout << "    Press any key to start the simulation and then any key to terminate.\n";
  FullScreen.use();
  setwrap(FALSE);
  textbackground(Black);
  textcolor(IsBW ? LightGray : Green);
  for (int r1 = 0; r1 < SCREENWIDTH; r1++)
    line[r1] = '';
  for (int r = 1; r<=SCREENDEPTH; r++)
    _directwrite(1,r,&line,SCREENWIDTH) ;
  textbackground(Black) ;
  srand((int)time(NULL));
  getch();
}

//#pragma restore

void Statistics()
{ window MsgW(5,5,37,10,SINGLEFRAME,IsBW ? Black : Blue,LightGray,
                                    IsBW ? Black : Red, LightGray);
  MsgW.settitle(" JPI TopSpeed C++ Demo ",CenterUpperTitle) ;
  unsigned X1 = 5,Y1 = 5,X2 = 37,Y2 = 10;
  MsgW.use();
  unsigned count = 50 ;
  unsigned X=X1,Y=Y1;
  while (1) {
    conout << "\nCars: " << Car::NumCars << "  Average Speed: " << AvSpeed/10
         << "." << AvSpeed%10;
    conout.flush();   // BUG WORKAROUND
    if (count==0) {
      X = randto(50);
      Y = randto(20);
      count = 200;
    }
    else
      count--;
    int changed = TRUE ;
    if      (X1+2<X) { X1+=3;X2+=3; }
    else if (X1>X+2) { X1-=3;X2-=3; }
    else if (Y1==Y) changed = FALSE ;
    if      (Y1<Y) { Y1++;Y2++; }
    else if (Y1>Y) { Y1--;Y2--; }
    if (changed)
    {
      MsgW.change(X1,Y1,X2,Y2);
      Delay(1);
    }
    else
      Delay(5);
  }
}

void main()
{
  Intro();
  InitWorld();
  AvSpeed = 300;
  StartScheduler();
  StartProcess(Statistics,0x2000,1);
  Car::Drive();
  gotoxy(1,SCREENDEPTH);
  cursoron();
}