/* Release 3.00 */

/*-------------------------------------------------------------------------*
*                                                                          *
*  task.cpp - task class implementations.      		                   *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved       *
*  Written by Simon Knight.                                                *
*                                                                          *
*--------------------------------------------------------------------------*/

#define _USE_TIME

#include <clib.h>
#include <task.hpp>

#pragma module(init_prio=>22)

#pragma save
extern "C" void (*_trans_adr(void (task::* )(), task&))();
#pragma call(reg_param=>(ax, bx, es))
extern "C" semaphore::sem_state set_sem(semaphore::sem_state , volatile semaphore::sem_state&  _sem);
#pragma restore

long get_tics() {

  struct time ts;
  time_t tt;

  time(&tt);
  gettime(&ts);
  return tt*1000 + ts.ti_hund*10;
}

//**************************************************************************
// task implementation
//**************************************************************************

task_control task::tc[MaxProcess];

task::task() {

  initialized=0;
}

void task::init(void (*func)(), unsigned stack_size, unsigned priority, int start) {

  thread_no=MakeReadyProcess(func, stack_size, priority, &np);
  initialized=1;
  tc[thread_no].Started=start;
  tc[thread_no].Destroyed=0;
  thread_active=start;
  if(start)
    StartReadyProcess(np);
  return;
}

task::~task() {

  tc[thread_no].Destroyed=1;
}

void task::start() {

  if((initialized) && (!tc[thread_no].Started)) {
    thread_active=1;
    tc[thread_no].Started=1;
    StartReadyProcess(np);
  }
}

void task::kill() {

  thread_active=0;
  tc[thread_no].Destroyed=1;
}

static void task::stop() {

  StopProcess();
  return;
}

static unsigned task::get_priority() {

  return CoreProc$CurrentPriority();
}

static void task::set_priority(unsigned p) {

  CoreProc$NewPriority(p);
}

void task::thread_proc() {

  while(!destroyed());
  stop();
}

void (* task::proc_adr())() {

//  return _trans_adr(&task::thread_proc, *this);
  return thread_proc;
}

static int task::destroyed() {

  return tc[_getTID()].Destroyed;
}

//**************************************************************************
// Semaphore implementation
//**************************************************************************

semaphore::semaphore() {

  sem=_clear;
}


semaphore::sem_state semaphore::set() {

  return set_sem(_set, sem);
}

semaphore::sem_state semaphore::clear() {

  return set_sem(_clear, sem);
}

semaphore::sem_state semaphore::wait(int msecs) {

  if(msecs == -1) {
    while(sem == _set) Delay(1);
    return _clear;
  }

  long target=get_tics()+msecs;
  while(sem == _clear) {
    if(get_tics() >= target)
      return _failed;
    Delay(1);
  }
  return _clear;
}

semaphore::sem_state semaphore::set_and_wait(int msecs) {

  sem = _set;
  Delay(0);

  if(msecs == -1) {
    while(sem == _set) Delay(1);
    return _clear;
  }

  long target=get_tics()+msecs;
  while(sem == _clear) {
    if(get_tics() >= target)
      return _failed;
    Delay(1);
  }
  return _clear;
}

semaphore::sem_state semaphore::request(int msecs) {

  if(msecs == -1) {
    while(set_sem(_set, sem) == _set) Delay(1);
    return _clear;
  }

  long target=get_tics()+msecs;
  while(set_sem(_set, sem) == _set) {
    if(get_tics() >= target)
      return _failed;
    Delay(1);
  }
  return _clear;
}

//**************************************************************************
// critical region implementation
//**************************************************************************

critical_region::critical_region() {

  s = new semaphore;
  alloc=1;
  n=0;
}

critical_region::critical_region(semaphore* sm) {

  s = sm;
  alloc=0;
  n=0;
}

critical_region::critical_region(critical_region& ) {

  InternalError(0x58);
}

critical_region::~critical_region() {

  if(alloc)
    delete s;
}

critical_region& critical_region::operator=(critical_region& ) {

  InternalError(0x58);
  return *this;
}

void critical_region::begin() {

   unsigned this_thread;

   this_thread = _getTID();
   if((n == 0) || (locked_thread != this_thread)) {
      s->request();
      locked_thread = this_thread;
   }
   ++n;
   return;
}

void critical_region::end() {

  if(--n <= 0) {
    n=0;
    s->clear();
    Delay(0);
  }
  return;
}

//**************************************************************************
// event implementation
//**************************************************************************

event::event() {

  s = new semaphore;
  s->set();
  count=0;
  alloc=1;
}

event::event(semaphore* sm) {

  s = sm;
  s->set();
  count=0;
  alloc=0;
}

event::event(event& ) {

  InternalError(0x58);
}

event::~event() {

  if(alloc)
    delete s;
}

event& event::operator=(event& ) {

  InternalError(0x58);
  return *this;
}

void event::wait() {

  cr.begin();
  if(--count < 0) {
    cr.end();
    s->wait();
  }
  else
    cr.end();
}

void event::trigger() {

  cr.begin();
  if(count++ < 0)
    s->clear();
  cr.end();
}

void event::notify() {

  cr.begin();
  if(count < 0) {
    s->clear();
    ++count;
  }
  cr.end();
}


void event::reset() {

  cr.begin();
  s->set();
  count=0;
  cr.end();
}


